/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

package ZEN_IDL.be;

import java.io.*;
import java.util.*;
import ZEN_IDL.ast.*;
import ZEN_IDL.be.helperUtils.*;
import util.*;

/*********************************************************
* Generates files <interfaceName>Helper.java and 
* <interfaceName>Operations.java.
*
* @author Olav Haugan
*
* 14.10.01
*********************************************************/
public class MQCCodeGenInterface
{
	private String interfaceName; /* The name of interface*/
	private InterfaceInheritence inheritence; /* The inheritence information */
	private Vector interfaceMethods; /* The interface methods */
	private String safeName; /* The full name of interface*/
	private VisitorState visitorState; /* The current state */
	private MQCCodeGenHelper codeGenHelper; /* code generation helper */
	private PrintWriter pw;

	private MQCConfig config;
		
	/*********************************************************
	* Constructor. Initializes variables.
	* 
	*********************************************************/
	public MQCCodeGenInterface(MQCConfig config)
	{
		codeGenHelper = new MQCCodeGenHelper(config);
		this.config = config;
		safeName = null;
		visitorState = null;
		pw = null;
	}	
	
	/*********************************************************
	* Sets the safe name variable
	* 
	* @param String The full name of the interface
	*
	*********************************************************/
	public void setSafeName(String safeName)
	{
		this.safeName = safeName;
	}

	/*********************************************************
	* Sets the interface name variable
	* 
	* @param String The name of the interface
	*
	*********************************************************/
	public void setInterfaceName(String interfaceName)
	{
		this.interfaceName = interfaceName;
	}

	/*********************************************************
	* Sets the inheritence variable
	* 
	* @param InterfaceInheritence The interface inheritence
	*
	*********************************************************/
	public void setInheritence(InterfaceInheritence inheritence)
	{
		this.inheritence = inheritence;
	}

	/*********************************************************
	* Sets the interface methods vector
	* 
	* @param Vector The interfaces methods
	*
	*********************************************************/
	public void setInterfaceMethods(Vector interfaceMethods)
	{
		this.interfaceMethods = interfaceMethods;
	}

	/*********************************************************
	* Sets the outputstream to write to
	* 
	* @param PrintWriter The outputstream
	*
	*********************************************************/
	public void setPW(PrintWriter pw)
	{
		this.pw = pw;
	}

	/*********************************************************
	* Gets the outputstream to write to
	* 
	* @param PrintWriter The outputstream
	*
	*********************************************************/
	public PrintWriter getPW()
	{
		return pw;
	}

	/*********************************************************
	* Sets the visitor state variable
	*
	* @param VisitorState The visitorState object
	*
	*********************************************************/
	public void setVisitorState(VisitorState visitorState)
	{
		this.visitorState = visitorState;
	}
	
	/*********************************************************
	* Generates the <interfaceName>Operations.java file.
	* This is the interface that the POA and the implementation
	* implements.
	*********************************************************/
	public void generateInterfaceOperations()
	{
		Enumeration e = interfaceMethods.elements();
		InterfaceMethod iMethod = null;
		Vector params = null;
		Enumeration ex = null;
		Vector elem = null;		
		
		while( e.hasMoreElements() ) 
		{
			iMethod = (InterfaceMethod) e.nextElement();
			params = iMethod.paramVector;
			pw.print("\tpublic " + codeGenHelper.mapIDL2Java( iMethod.returnType ) + " " + iMethod.name + "( " );
			codeGenHelper.printParams(pw, params);
			pw.print( " )");

			ex = iMethod.exceptionVector.elements();
			if(ex.hasMoreElements())
			{
				pw.print( " throws mqc.exception.UNKNOWN, ");
				while(ex.hasMoreElements())
				{
					elem = (Vector) ex.nextElement();
					Enumeration exceptionElement = elem.elements();					
					while(exceptionElement.hasMoreElements())
					{
						pw.print(exceptionElement.nextElement().toString());
						if(exceptionElement.hasMoreElements())
							pw.print(".");
					}
					if(ex.hasMoreElements())
						pw.print(", ");
				}
				pw.println( ";");
			}
			pw.println( ";");
		}
	}
	
	/*********************************************************
	* Generates the start of the 
	* <interfaceName>Operations.java file.
	*********************************************************/
	public void genStartOperations()
	{
		String packagePath = visitorState.getPackagePath();
		String packageStr = visitorState.getPackageString();

		PrintWriter pw = MQCCodeGenHelper.makeFile(packagePath, interfaceName + "Operations");
		setPW(pw);
		
		if(packagePath.length() != 0)
			pw.println(packageStr);
		pw.println();
		pw.print("public interface " + interfaceName + "Operations");
	
		String inheritClassName = null; 
		String inheritPackageName = null;
		Vector inheritenceSpec = inheritence.getInheritence();

		Enumeration e = inheritenceSpec.elements();
			
		if( e.hasMoreElements() )
			pw.print(" extends ");
		
		Vector nSpec = null;
		while( e.hasMoreElements() ) 
		{
			nSpec = (Vector) e.nextElement();
			inheritClassName = InterfaceInheritence.getInterfaceName( nSpec );
			inheritPackageName = InterfaceInheritence.getPackage( nSpec );

			if( visitorState.isInterfaceAbstract( inheritClassName , inheritPackageName ) )
				pw.print( inheritClassName );
			else
				pw.print( inheritClassName + "Operations");
			if( e.hasMoreElements() )
				pw.print(", " );
		}
		pw.println();
		pw.println("{");
	}

	/*********************************************************
	* Generates the <interface>Helper.java file
	* <interfaceName>Operations.java file.
	*********************************************************/
	public void generateInterfaceHelper()
	{
		String packagePath = visitorState.getPackagePath();
		String packageStr = visitorState.getPackageString();

		PrintWriter pw = MQCCodeGenHelper.makeFile(packagePath, interfaceName + "Helper");
		if(packagePath.length() != 0)
			pw.println(packageStr);
		pw.println();
		
		pw.println("public final class " + interfaceName + "Helper");
		pw.println("{");

		pw.println("\tprivate static final String ID = \"" + codeGenHelper.generateID(packagePath, safeName) + "\";");		
		pw.println();
		printGetID(pw);
		pw.println();
		printNarrow(pw);
		pw.println("};");
		pw.println("// End of file");
		pw.flush();
		pw.close();
	}
	
	private void printGetID(PrintWriter pw)
	{
		pw.println("\tpublic static String id()");
		pw.println("\t{");
		pw.println("\t\treturn ID;");
		pw.println("\t}");
	}

	private void printNarrow(PrintWriter pw)
	{
		String packagePath = visitorState.getPackagePath();
		String fullName = codeGenHelper.generateFullObjName( packagePath, interfaceName );
		String stubName = codeGenHelper.generateFullObjName( packagePath, "_" + interfaceName + "Stub" );

		pw.println("\tpublic static " + fullName + " narrow( final Object object )");
		pw.println("\t{");
		pw.println("\t\tif( object == null)");
		pw.println("\t\t{");
		pw.println("\t\t\treturn null;");
		pw.println("\t\t}");
		pw.println("\t\tif( object instanceof " + fullName + " )");
		pw.println("\t\t{");
		pw.println("\t\t\treturn ( " + fullName + " ) object;");
		pw.println("\t\t}");
		pw.println("\t\tmqc.Delegate delegate = ((mqc.ObjImpl) object).get_delegate();");
		//String protocol = config.getClientProtocol();
		//pw.println("\t\tmqc.Delegate delegate = new mqc." + protocol + "Delegate(dlg);");
		String transport = config.getClientTransport();		
		//pw.println("\t\tmqc.transports.Transport transport = new mqc.transports." + transport + "Transport(delegate);");
		//pw.println("\t\tdelegate.set_transport(transport);");
		// here we also need to extract any secondary transports that may have been
		//	selected for spatial redundancy
		Hashtable htSpatial = config.getFTSpatial();
		String strFTSTCP = (String) htSpatial.get("ftsTCP");
		String strFTSUnUDP = (String) htSpatial.get("ftsUnUDP");
		String strFTSGBNUDP = (String) htSpatial.get("ftsGBNUDP");
		String strFTSSNWUDP = (String) htSpatial.get("ftsSNWUDP");
		String strFTSSerial = (String) htSpatial.get("ftsSerial");
		String strFTSOneWire = (String) htSpatial.get("ftsOneWire");
		int intTransportNum = 0;
		if(transport.equals("TCP") || strFTSTCP.equals("true")) {
			pw.println("\t\tmqc.transports.Transport transport" + Integer.toString(intTransportNum) + " = new mqc.transports." + Constants.TR_TCP_STR + "Transport(delegate);");
			intTransportNum++;
		}
		if ( transport.equals(Constants.TR_UNRELIABLE_UDP_STR) || strFTSUnUDP.equals("true")) {
			pw.println("\t\tmqc.transports.Transport transport" + Integer.toString(intTransportNum) + " = new mqc.transports." + Constants.TR_UNRELIABLE_UDP_STR + "Transport(delegate);");
			intTransportNum++;
		} else if ( transport.equals(Constants.TR_GOBACKN_UDP_STR) || strFTSGBNUDP.equals("true")) {
			pw.println("\t\tmqc.transports.Transport transport" + Integer.toString(intTransportNum) + " = new mqc.transports." + Constants.TR_GOBACKN_UDP_STR + "Transport(delegate);");
			intTransportNum++;
		} else if ( transport.equals(Constants.TR_STOPANDWAIT_UDP_STR) || strFTSSNWUDP.equals("true")) {
			pw.println("\t\tmqc.transports.Transport transport" + Integer.toString(intTransportNum) + " = new mqc.transports." + Constants.TR_STOPANDWAIT_UDP_STR + "Transport(delegate);");
			intTransportNum++;
		}
		//pw.println("\t\tMQC.clientTransports.Transport transport = new MQC.clientTransports." + transport + "Transport(delegate);");
		//pw.println("\t\tdelegate.set_transport(transport);");
		for(int i=0;i<intTransportNum;i++) {
			pw.println("\t\tdelegate.add_transport(transport" + Integer.toString(i) + ");");
		}
		pw.println("\t\t" + stubName + " stub = new " + stubName + "();" );
		pw.println("\t\tstub.set_delegate(delegate);");
		pw.println("\t\treturn stub;");
//		pw.println("\t\tthrow new mqc.exception.BAD_PARAM();");
		pw.println("\t}");
	}

	/*********************************************************
	* Generates the <interfaceName>.java file.
	*********************************************************/
	public void generateInterface()
	{
		String packagePath = visitorState.getPackagePath();
		String packageStr = visitorState.getPackageString();
		String fullName = codeGenHelper.generateFullObjName( packagePath, interfaceName );

		PrintWriter pw = MQCCodeGenHelper.makeFile(packagePath, interfaceName);
		
		if(packagePath.length() != 0)
			pw.println(packageStr);
		pw.println();
		pw.println("public interface " + interfaceName + " extends " + fullName + "Operations");
		pw.println("{");
		pw.println("};");
		pw.println("// End of file");		
		pw.flush();
		pw.close();
	}

}
