/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

package util;

import java.io.*;
import java.util.*;
import java.lang.*;
import ZEN_IDL.be.*;

/*********************************************************
* Class to load the configuration properties used for
* constraint checking and generating code.
*
* @author Olav Haugan
* @author Tarana Damania
* @author David McKinnon
* @author Thor Egil Skaug
*********************************************************/
public class MQCConfig
{
  private Properties props; //The configuration properties.
  private HashMap dataTypes; //boolean values for data types.
  private boolean padding; //Whether to pad or not
    
  /*********************************************************
  * Constructor. Loads the properties from file and
  * populates the data types hash map.
  *
  * @param String The name of the properties file
  *********************************************************/
  public MQCConfig(String fileName)
  {
    this.padding = true;
    props = new Properties();
    try
    {
      props.load(new FileInputStream(fileName));
    }
    catch(IOException ex)
    {
      System.out.println("Error loading properties file " + fileName);
      System.out.println("Please specify a valid properties file");
      System.exit(-1);
    }
    dataTypes = new HashMap(Constants.NO_PRIMITIVE_TYPES+
                            Constants.NO_COMPLEX_TYPES+
                            Constants.NO_EXCEPTION_TYPES);
    updateDataTypes();
  }
  
  /*********************************************************
  * Constructor. Loads the properties from file and
  * populates the data types hash map.
  *
  * @param File The file with properties
  *********************************************************/
  public MQCConfig(File file)
  {
    this.padding = true;
    props = new Properties();
    try
    {
      props.load(new FileInputStream(file));
    }
    catch(IOException ex)
    {
      System.out.println("Error loading properties file " + file.getName());
      System.out.println("Please specify a valid properties file");
      System.exit(-1);
    }
    dataTypes = new HashMap(Constants.NO_PRIMITIVE_TYPES+
                            Constants.NO_COMPLEX_TYPES+
                            Constants.NO_EXCEPTION_TYPES);
    updateDataTypes();
  }

  /*********************************************************
  * Constructor. Initializes variables,
  * populates the data types hash map.
  *
  * @param Properties The properties
  *********************************************************/
  public MQCConfig(Properties props)
  {
    this.padding = true;
    this.props = props;

    dataTypes = new HashMap(Constants.NO_PRIMITIVE_TYPES+
                            Constants.NO_COMPLEX_TYPES+
                            Constants.NO_EXCEPTION_TYPES);
    updateDataTypes();
  }

  /*********************************************************
  * Populates the hash map with values from properties.
  *********************************************************/
  public void updateDataTypes()
  {
    int i = 0;
  
    dataTypes.put("void", "true");
    for(i = 0; i < Constants.NO_PRIMITIVE_TYPES; i++)
    {
      if(getBooleanProperty(Constants.PRIMITIVE_TYPE[i]))
        dataTypes.put(Constants.PRIMITIVE_TYPE[i], "true");
    }
    for(i = 0; i < Constants.NO_COMPLEX_TYPES; i++)
    {
      if(getBooleanProperty(Constants.COMPLEX_TYPE[i]))
        dataTypes.put(Constants.COMPLEX_TYPE[i], "true");
    }
    for(i = 0; i < Constants.NO_EXCEPTION_TYPES; i++)
    {
      if(getBooleanProperty(Constants.EXCEPTION_TYPE[i]))
        dataTypes.put(Constants.EXCEPTION_TYPE[i], "true");
    }
  }

  /*********************************************************
  * Checks if a data type is allowed or not.
  *
  * @param String The name of the data type
  *
  * @return boolean true if a valid data type, false
  *           otherwise.
  *********************************************************/
  public boolean isValidDataType(String type)
  {
    String valid = (String) dataTypes.get(MQCTypedefMapping.get(type));
    return (valid != null);
  }

  /*********************************************************
  * Checks if system exceptions is selected.
  *
  * @return boolean true if system exceptions is 
  *          selected, false otherwise.
  *********************************************************/
  public boolean isSystemException()
  {
    String valid = (String) dataTypes.get(Constants.EXCEPTION_TYPE[0]);
    return (valid != null);
  }

  /*********************************************************
  * Checks if user exceptions is selected.
  *
  * @return boolean true if user exceptions is 
  *          selected, false otherwise.
  *********************************************************/
  public boolean isUserException()
  {
    String valid = (String) dataTypes.get(Constants.EXCEPTION_TYPE[1]);
    return (valid != null);
  }

  /*********************************************************
  * Checks if we have a valid number of interfaces.
  *
  * @param int The number of interfaces.
  *
  * @return boolean true if a valid, false
  *           otherwise.
  *********************************************************/
  public boolean isValidNoInterfaces(int value)
  {
    int cfgValue = getIntProperty(Constants.NO_INTERFACES);
    if(cfgValue == 0)
      return true; //no restriction
      
    return !(cfgValue < value);
  }

  /*********************************************************
  * Returns the number of interfaces allowed
  *
  * @return int The number of interfaces allowed
  *********************************************************/
  public int getNoInterfaces()
  {
    return getIntProperty(Constants.NO_INTERFACES);
  }

  /*********************************************************
  * Returns the value of padding variable.
  *
  * @return boolean true if we should pad, false otherwise.
  *********************************************************/
  public boolean isPadding()
  {
    return padding;
  }

  /*********************************************************
  * Sets the value of padding variable.
  *
  * @param boolean The padding value
  *********************************************************/
  public void setPadding(boolean padding)
  {
    this.padding = padding;
  }

  /*********************************************************
  * Checks if we have a valid number of methods.
  *
  * @param int The number of methods.
  *
  * @return boolean true if a valid, false
  *           otherwise.
  *********************************************************/
  public boolean isValidNoMethods(int value)
  {
    int cfgValue = getIntProperty(Constants.NO_METHODS);
    if(cfgValue == 0)
      return true; //no restriction

    return !(cfgValue < value);
  }

  /*********************************************************
  * Returns the number of methods allowed
  *
  * @return int The number of methods allowed
  *********************************************************/
  public int getNoMethods()
  {
    return getIntProperty(Constants.NO_METHODS);
  }

  /*********************************************************
  * Checks if we have a valid payload size.
  *
  * @param int The payload size
  *
  * @return boolean true if a valid, false
  *           otherwise.
  *********************************************************/
  public boolean isValidMSGPayload(int value)
  {
    int cfgValue = getIntProperty(Constants.MSG_PAYLOAD);
    if(cfgValue == 0)
      return true; //no restriction

    return !(cfgValue < value);
  }

  /*********************************************************
  * Checks if we should produce inline marshaling/
  * un-marshaling or not.
  *
  * @return boolean true if a inline, false
  *           otherwise.
  *********************************************************/
  public boolean isInline()
  {
    return getBooleanProperty(Constants.INLINE_MARSHAL);
  }

  /*********************************************************
  * Checks if we have homogeneous hw, or not
  *
  * @return boolean true if a homogeneous, false
  *           otherwise.
  *********************************************************/
  public boolean isHomogeneousHW()
  {
    return getBooleanProperty(Constants.HOMOGENEOUS_HW);
  }

  /*********************************************************
  * Checks if we are forcing endianess
  *
  * @return boolean true if a homogeneous, false
  *           otherwise.
  *********************************************************/
  public int isForcingEndianess()
  {
    if(getBooleanProperty(Constants.FORCE_ENDIANESS))
    {
      return (getIntProperty(Constants.ENDIANESS)+1);
    }
    else
      return 0;
  }

  /*********************************************************
  * Checks if we have Fault tolerant temporal properties
  *
  * @return String FT Temporal value
  *********************************************************/
  public String getFTTemporal() {
    return props.getProperty("temporal");
  }
  
  /*********************************************************
  * Checks if we have FT Spatial properties and
  * returns a Hashtable with the settings
  *
  * @returns Hashtable with the FT Spatial settings
  *********************************************************/
  public Hashtable getFTSpatial() {
    Hashtable htSpatial = new Hashtable();
    // load the spatial properties into the hashtable
    htSpatial.put("enableSpatial",props.getProperty("enableSpatial"));
    htSpatial.put("ftsTCP",props.getProperty("ftsTCP"));
    htSpatial.put("ftsUnUDP",props.getProperty("ftsUnUDP"));
    htSpatial.put("ftsGBNUDP",props.getProperty("ftsGBNUDP"));
    htSpatial.put("ftsSNWUDP",props.getProperty("ftsSNWUDP"));
    htSpatial.put("ftsSerial",props.getProperty("ftsSerial"));
    htSpatial.put("ftsOneWire",props.getProperty("ftsOneWire"));
    // now return the hashtable
    return htSpatial;
  }
        
        /*********************************************************
  * Checks if we have FT GroupCom properties and
  * returns a Hashtable with the settings
  *
  * @returns Hashtable with the FT GroupCom settings
  *********************************************************/
  public Hashtable getFTGroupCom() {
    Hashtable htGroupCom = new Hashtable();
    // load the spatial properties into the hashtable
    htGroupCom.put("enableGroupCom",props.getProperty("enableGroupCom"));
    htGroupCom.put("ftGCUniform",props.getProperty("ftGCUniform"));
    htGroupCom.put("ftGCOrder",props.getProperty("ftGCOrder"));
    htGroupCom.put("ftGCTimeOut",props.getProperty("ftGCTimeOut"));
    htGroupCom.put("ftGCTrans",props.getProperty("ftGCTrans"));
    // now return the hashtable
    return htGroupCom;
  }
        
  /*********************************************************
  * Checks if we have FT Value redundancy properties and
  * returns a String with the setting
  *
  * @returns String with the FT Value Redundancy setting
  *********************************************************/
  public String getFTValueRed() {
    return props.getProperty("enableValue");
  }
        
  /*********************************************************
  * Checks if we have Security properties set
  *
  * @return boolean true if security is selected, false
  *                   otherwise.
  *********************************************************/
  public boolean getSecurityCipherStatus()
  {
    return getBooleanProperty(Constants.SECURITY);
  }

  /*********************************************************
  * Check to see which (if any) security cipher is selected
  *
  * @return String The security cipher specification
  *********************************************************/
  public String getSecurityCipher()
  {
    String secCipher = null;  
    if(getBooleanProperty(Constants.SECURITY)) {
      secCipher = props.getProperty(Constants.SEC_CIPHER);
    }
    return secCipher;
  }

  /*********************************************************
  * Check to see which (if any) security cipher mode is selected
  *
  * @return String The security cipher specification
  *********************************************************/
  public String getSecurityCipherMode()
  {
    String secCipherMode = null;  
    if(getBooleanProperty(Constants.SECURITY)) {
      secCipherMode = props.getProperty(Constants.SEC_CIPHER_MODE);
    }
    return secCipherMode;
  }

  /*********************************************************
  * Check to see which (if any) security cipher mode is selected
  *
  * @return String The security cipher specification
  *********************************************************/
  public String getSecurityCipherPadding()
  {
    String secCipherPadding = null;  
    if(getBooleanProperty(Constants.SECURITY)) {
      secCipherPadding = props.getProperty(Constants.SEC_CIPHER_PADDING);
    }
    return secCipherPadding;
  }

  /*********************************************************
  * Check to see what the cipher key length is set to
  *
  * @return int cipher key length (in bits)
  *********************************************************/
  public int getSecurityCipherKeyLen()
  {
    int secCipherKeyLen = 0;  
    if(getBooleanProperty(Constants.SECURITY)) {
      secCipherKeyLen = getIntProperty(Constants.SEC_CIPHER_KEY_LEN);
    }
    return secCipherKeyLen;
  }

  /*********************************************************
  * Checks if we have Security Message Digest properties set
  *
  * @return boolean true if security MD is selected, false otherwise.
  *********************************************************/
  public boolean getSecurityMDStatus()
  {
    return getBooleanProperty(Constants.SECURITY_MD);
  }

  /*********************************************************
  * Check to see which (if any) security message digest is selected
  *
  * @return String The security MD specification
  *********************************************************/
  public String getSecurityMD()
  {
    String secMD = null;  
    if(getBooleanProperty(Constants.SECURITY_MD)) {
      secMD = props.getProperty(Constants.SEC_MD_LABEL);
    }
    return secMD;
  }
  
  /*********************************************************
  * Checks if we have Security Message Digest is not being used as an HMAC
  *
  * @return boolean true if MD is not being used for an HMAC, false otherwise.
  *********************************************************/
  public boolean getSecurityMDStatusNotHMAC()
  {
    if (getSecurityMACStatus()) {
        if (getSecurityMAC().equalsIgnoreCase("HMAC")) { 
          return false;
        }
    }
    return true;
  }

  /*********************************************************
  * Checks if we have Security Message Authentication Code properties set
  *
  * @return boolean true if security MAC is selected, false otherwise.
  *********************************************************/
  public boolean getSecurityMACStatus()
  {
    return getBooleanProperty(Constants.SECURITY_MAC);
  }

  /*********************************************************
  * Check to see which (if any) Security MAC is selected
  *
  * @return String The security MAC specification
  *********************************************************/
  public String getSecurityMAC()
  {
    String secMAC = null;  
    if(getBooleanProperty(Constants.SECURITY_MAC)) {
      secMAC = props.getProperty(Constants.SEC_MAC_LABEL);
    }
    return secMAC;
  }

  /*********************************************************
  * Check to see what the MAC key length is set to
  *
  * @return int cipher key length (in bits)
  *********************************************************/
  public int getSecurityMACKeyLen()
  {
    int secMACKeyLen = 0;  
    if(getBooleanProperty(Constants.SECURITY_MAC)) {
      secMACKeyLen = getIntProperty(Constants.SEC_MAC_KEY_LEN);
    }
    return secMACKeyLen;
  }

  /*********************************************************
  * Checks if we need to support security keys within the ORB
  *
  * @return boolean true if ORB keys are needed, false otherwise.
  *********************************************************/
  public boolean getSecurityORBKeyStatus()
  {
    return getSecurityCipherStatus() || getSecurityMACStatus();
  }

  /*********************************************************
  * Checks if we have Security Audit properties set
  *
  * @return boolean true if Security Audit is selected, false otherwise.
  *********************************************************/
  public boolean getSecurityAuditStatus()
  {
    return getBooleanProperty(Constants.SECURITY_AUDIT);
  }

  /*********************************************************
  * Check to see which (if any) Security Audit alg. is selected
  *
  * @return String The security MAC specification
  *********************************************************/
  public String getSecurityAudit()
  {
    String secAudit = null;  
    if(getBooleanProperty(Constants.SECURITY_AUDIT)) {
      secAudit = props.getProperty(Constants.SEC_AUDIT_LABEL);
    }
    return secAudit;
  }

  /*********************************************************
  * Check to see what the Audit Log length is set to
  *
  * @return int cipher key length (in bits)
  *********************************************************/
  public int getSecurityAuditLogLen()
  {
    int secAuditLogLen = 0;  
    if(getBooleanProperty(Constants.SECURITY_AUDIT)) {
      secAuditLogLen = getIntProperty(Constants.SEC_AUDIT_LOG_LEN);
    }
    return secAuditLogLen;
  }

  /*********************************************************
  * Checks if we have Real time properties and 
  * returns a vector of Real time property names.
  *
  * @return Vector A vector of Real time config. names
  *********************************************************/
  public Vector getRealTime()
  {
    Vector rt = null;  
    if(getBooleanProperty(Constants.REALTIME))
    {
      if(getBooleanProperty(Constants.RT_SOFT))
      {
        rt.add(Constants.RT_SOFT);
      }
      if(getBooleanProperty(Constants.RT_HARD))
      {
        rt.add(Constants.RT_HARD);
      }
    }
    return rt;
  }

  /*********************************************************
  * Checks what type of transport the client is using
  *
  * @return String The name of the transport used
  *********************************************************/
  public String getClientTransport()
  {
    if(getBooleanProperty(Constants.TR_SERIAL))
    {
      return Constants.TR_SERIAL_STR;
    }
    else if(getBooleanProperty(Constants.TR_1WIRE))
    {
      return Constants.TR_1WIRE_STR;
    }
    else if(getBooleanProperty(Constants.TR_TCP))
    {
      return Constants.TR_TCP_STR;
    }
    else if(getBooleanProperty(Constants.TR_UNRELIABLE_UDP))
    {
      return Constants.TR_UNRELIABLE_UDP_STR;
    }
    else if(getBooleanProperty(Constants.TR_GOBACKN_UDP))
    {
      return Constants.TR_GOBACKN_UDP_STR;
    }
    else if(getBooleanProperty(Constants.TR_STOPANDWAIT_UDP))
    {
      return Constants.TR_STOPANDWAIT_UDP_STR;
    }

    return null;
  }

  /*********************************************************
  * Returns what type of protocol the client is using
  *
  * @return String The name of the protocol used
  *********************************************************/
  public String getClientProtocol()
  {
    if(getBooleanProperty(Constants.GIOP_GIOP))
    {
      return Constants.GIOP_GIOP_STR;
    }
    else if(getBooleanProperty(Constants.GIOP_LITE))
    {
      return Constants.GIOP_LITE_STR;
    }
    else if(getBooleanProperty(Constants.GIOP_MQC))
    {
      return Constants.GIOP_MQC_STR;
    }
    return null;
  }

  /*********************************************************
  * Checks what type of transports the server is using
  *
  * @return Vector A list of transports
  *********************************************************/
  public Vector getServerTransport()
  {
    Vector transports = new Vector(1,1);

    if(getBooleanProperty(Constants.SERVER_HDR + "_" + Constants.TR_SERIAL))
    {
      transports.add(Constants.TR_SERIAL_STR);
    }
    if(getBooleanProperty(Constants.SERVER_HDR + "_" + Constants.TR_1WIRE))
    {
      transports.add(Constants.TR_1WIRE_STR);
    }
    if(getBooleanProperty(Constants.SERVER_HDR + "_" + Constants.TR_TCP))
    {
      transports.add(Constants.TR_TCP_STR);
    }
    if(getBooleanProperty(Constants.SERVER_HDR + "_" + Constants.TR_UNRELIABLE_UDP))
    {
      transports.add(Constants.TR_UNRELIABLE_UDP_STR);
    }
    if(getBooleanProperty(Constants.SERVER_HDR + "_" + Constants.TR_GOBACKN_UDP))
    {
      transports.add(Constants.TR_GOBACKN_UDP_STR);
    }
    if(getBooleanProperty(Constants.SERVER_HDR + "_" + Constants.TR_STOPANDWAIT_UDP))
    {
      transports.add(Constants.TR_STOPANDWAIT_UDP_STR);
    }
    return transports;
  }

  /*********************************************************
  * Returns what type of protocols the server is using
  *
  * @return Vector A list of protocols used.
  *********************************************************/
  public Vector getServerProtocol()
  {
    Vector protocols = new Vector(1,1);
    
    if(getBooleanProperty(Constants.SERVER_HDR + "_" + Constants.GIOP_GIOP))
    {
      protocols.add(Constants.GIOP_GIOP_STR);
    }
    if(getBooleanProperty(Constants.SERVER_HDR + "_" +  Constants.GIOP_LITE))
    {
      protocols.add(Constants.GIOP_LITE_STR);
    }
    if(getBooleanProperty(Constants.SERVER_HDR + "_" + Constants.GIOP_MQC))
    {
      protocols.add(Constants.GIOP_MQC_STR);
    }
    return protocols;
  }

  /*********************************************************
  * Checks if we have Network QoS properties and 
  * returns a vector of Network QoS property names.
  *
  * @return Vector A vector of Net. QoS config. names
  *********************************************************/
  public Vector getNetQoS()
  {
    Vector netQoS = null;  
    if(getBooleanProperty(Constants.NETWORK))
    {
      if(getBooleanProperty(Constants.NET_GUARANTEED_BW))
      {
        netQoS.add(Constants.NET_GUARANTEED_BW);
      }
      if(getBooleanProperty(Constants.NET_LOW_LATENCY))
      {
        netQoS.add(Constants.NET_LOW_LATENCY);
      }
    }
    return netQoS;
  }

  /*********************************************************
  * Returns the bandwidth requirement
  *
  * @return int The bandwidth requirement
  *********************************************************/
  public int getBandWidth()
  {
    return getIntProperty(Constants.NET_GUARANTEED_BW_NUM);
  }

  /*********************************************************
  * Checks if we are compressing interface names/ method
  * names.
  *
  * @return boolean true if compressing, false otherwise
  *********************************************************/
  public boolean isCompressing()
  {
    return getBooleanProperty(Constants.COMPRESS_NAMES);
  }

  /*********************************************************
  * Determine the Debug Status (true/false).
  *
  * @return boolean true if DEBUG is on, false otherwise
  *********************************************************/
  public boolean getDebugStatus()
  {
    return getBooleanProperty(Constants.DEBUG);
  }

  /*********************************************************
  * Determine the Debug Level.
  *
  * @return int which is the current debug level
  *********************************************************/
  public int getDebugLevel()
  {
    return getIntProperty(Constants.DEBUG_LEVEL);
  }

  /*********************************************************
  * Determine the Debug Memory Status (true/false).
  *
  * @return boolean true if DEBUG_MEMORY is on, false otherwise
  *********************************************************/
  public boolean getDebugMemoryStatus()
  {
    return getBooleanProperty(Constants.DEBUG_MEMORY);
  }

  /*********************************************************
  * Determine the Debug Timing Status (true/false).
  *
  * @return boolean true if DEBUG_TIMING is on, false otherwise
  *********************************************************/
  public boolean getDebugTimingStatus()
  {
    return getBooleanProperty(Constants.DEBUG_TIMING);
  }

  /*********************************************************
  * Determine if Java CLDC support is needed.
  *
  * @return int which is the current debug level
  *********************************************************/
  public boolean getCLDCStatus()
  {
    return getBooleanProperty(Constants.FORCE_CLDC);
  }

  /*********************************************************
  * Gets an integer property from the property configuration.
  *
  * @param String The name of the property.
  *
  * @return int The property value.
  *********************************************************/
  private int getIntProperty(String property)
  {
    return (new Integer(props.getProperty(property))).intValue();
  }

  /*********************************************************
  * Gets an boolean property from the property configuration.
  *
  * @param String The name of the property.
  *
  * @return boolean The property value.
  *********************************************************/
  private boolean getBooleanProperty(String property)
  {
    return (new Boolean(props.getProperty(property))).booleanValue();
  }
  
   public boolean getUnreliableUDPStatus()
   {
      if(getBooleanProperty(Constants.SERVER_HDR + "_" + Constants.TR_UNRELIABLE_UDP))
      {
         return true;
      }
      else if(getBooleanProperty(Constants.TR_UNRELIABLE_UDP))
      {
         return true;
      }
      return false;   
   }

   public boolean getGoBackNUDPStatus()
   {
      if(getBooleanProperty(Constants.SERVER_HDR + "_" + Constants.TR_GOBACKN_UDP))
      {
         return true;
      }
      else if(getBooleanProperty(Constants.TR_GOBACKN_UDP))
      {
         return true;
      }
      return false;
   }
  
   public boolean getStopNWaitUDPStatus()
   {
      if(getBooleanProperty(Constants.SERVER_HDR + "_" + Constants.TR_STOPANDWAIT_UDP))
      {
         return true;
      }
      else if(getBooleanProperty(Constants.TR_STOPANDWAIT_UDP))
      {
         return true;
      }
      return false;
   }
  /*********************************************************
  * Gets an boolean property from the property configuration.
  *
  * @param String The name of the property.
  *
  * @return boolean The property value.
  *********************************************************/
   public boolean getUDPStatus()
   {
      return (getUnreliableUDPStatus() || getGoBackNUDPStatus() || getStopNWaitUDPStatus());
/*    if(getBooleanProperty(Constants.SERVER_HDR + "_" + Constants.TR_UNRELIABLE_UDP))
    {
      return true;
    }
    else if(getBooleanProperty(Constants.SERVER_HDR + "_" + Constants.TR_GOBACKN_UDP))
    {
      return true;
    }
    else if(getBooleanProperty(Constants.SERVER_HDR + "_" + Constants.TR_STOPANDWAIT_UDP))
    {
      return true;
    }
    else if(getBooleanProperty(Constants.TR_UNRELIABLE_UDP))
    {
      return true;
    }
    else if(getBooleanProperty(Constants.TR_GOBACKN_UDP))
    {
      return true;
    }
    else if(getBooleanProperty(Constants.TR_STOPANDWAIT_UDP))
    {
      return true;
    }
    return false;*/
   } 
  
  /*********************************************************
  * Returns the packet size (based on max payload)
  * @return int that is the max payload (0 is unlimited)
  *********************************************************/
  public int getMSGPayload() {
     return getIntProperty(Constants.MSG_PAYLOAD);
  }

  /*********************************************************
  * Returns the packet size (based on max payload)
  * @return int that is the max payload (0 is unlimited)
  *********************************************************/
  public boolean supportsGIOPVersion(String version) {
     return getBooleanProperty(version);
  }

}
