/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

package util;

import java.util.*;
import java.io.*;

/*********************************************************
* Class for checking if the configuration options selected
* make sense.
*
* @author Olav Haugan
*
* Date: 14.10.01
*********************************************************/
public class ProfileChecker
{
	private Properties props;
	private Vector error;
	private MQCConfig config;

	public static final String HW_ERROR1 = "Cannot force endianess and use heterogeneous hardware.";
	public static final String HW_ERROR2 = "Client cannot use GIOPLite protocol and use heterogeneous hardware.";
	public static final String HW_ERROR3 = "Server cannot use GIOPLite protocol and use heterogeneous hardware.";
	public static final String PROTOCOL_ERROR1 = "Please select at least one protocol for server.";
	public static final String PROTOCOL_ERROR2 = "Client and server unable to communicate. No matching protocol for client and server.";
	public static final String TRANSPORT_ERROR1 = "Please select at least one transport for server.";
	public static final String TRANSPORT_ERROR2 = "Client and server unable to communicate. No matching transport for client and server.";
	public static final String EXCEPTION_ERROR1 = "Cannot use user exceptions without system exceptions.";

	/*********************************************************
	* Constructor. Initializes variables
	*
	* @param File The file containing the configuration opt.
	*********************************************************/
	public ProfileChecker( File file )
	{
		this.props =  null;
		config = new MQCConfig(file);

		this.error = new Vector(10,10);
	}

	/*********************************************************
	* Constructor. Initializes variables
	*
	* @param Properties The properties containing configuration
	*********************************************************/
	public ProfileChecker( Properties props )
	{
		this.props =  props;
		config = new MQCConfig(props);
		this.error = new Vector(10,10);
	}
	
	/*********************************************************
	* Checks the properties for impossible configurations.
	*
	* @return boolean true if the profile is valid, false
	*                  otherwise.
	*********************************************************/
	public boolean isProfileValid()
	{
		boolean valid = true;
		
		if(!checkDataTypes())
			valid = false;
			
		if(!checkExceptions())
			valid = false;
			
		if(!checkTransport())
			valid = false;
		
		if(!checkProtocols())
			valid = false;
		
		if(!checkHWConfig())
			valid = false;

		return valid;
	}
	
	/*********************************************************
	* Prints the configuration errors to screen
	*********************************************************/
	public void errors()
	{
		Enumeration e = error.elements();
		while(e.hasMoreElements())
		{
			System.out.println((String) e.nextElement());
		}
	}

	/*********************************************************
	* Return the error vector.
	*
	* @return Vector The list of errors
	*
	*********************************************************/
	public Vector getErrors()
	{
		return error;
	}
	
	/*********************************************************
	* Checks the data types for errors. 
	*
	* @return boolean true if errors, false otherwise
	*********************************************************/
	public boolean checkDataTypes()
	{
		return true;
	}

	/*********************************************************
	* Checks the HW config for errors. Error is produced if
	* endianess is forced and heterogeneous hw. is selected.
	* Error is also produced if heterogeneous hw. is selected
	* and GIOPLite is selected.
	*
	* @return boolean true if errors, false otherwise
	*********************************************************/
	public boolean checkHWConfig()
	{
		boolean valid = true;
		if((config.isForcingEndianess() > 0) & !config.isHomogeneousHW())
		{
			error.add(HW_ERROR1);
			valid = false;
		}
		String clientProtocol = config.getClientProtocol();
		if(clientProtocol.equals("GIOPLite") && !config.isHomogeneousHW())
		{
			error.add(HW_ERROR2);
			valid = false;
		}

		if(!config.isHomogeneousHW())
		{
			Vector protocols = config.getServerProtocol();
			if(protocols.size() == 1)
			{	
				Enumeration e = protocols.elements();
				while(e.hasMoreElements())
				{
					if(((String) e.nextElement()).equals("GIOPLite"))
					{
						error.add(HW_ERROR3);
						valid = false;
						break;
					}	
				}
			}
		}
		return valid;
	}

	/*********************************************************
	* Checks the selected protocols for errors. 
	*
	* @return boolean true if errors, false otherwise
	*********************************************************/
	public boolean checkProtocols()
	{
		boolean valid = true;
		
		String clientProtocol = config.getClientProtocol();
		Vector protocols = config.getServerProtocol();
		
		Enumeration e = protocols.elements();
		if(!e.hasMoreElements())
		{
			error.add(PROTOCOL_ERROR1);
			valid = false;
		}
		boolean match = false;
		while(e.hasMoreElements())
		{
			if(((String) e.nextElement()).equals(clientProtocol))
			{
				match = true;
				break;
			}	
		}
		if(!match)
		{
			error.add(PROTOCOL_ERROR2);
			valid = false;
		}
		return valid;
	}

	/*********************************************************
	* Checks the selected transports for errors. 
	*
	* @return boolean true if errors, false otherwise
	*********************************************************/
	public boolean checkTransport()
	{
		boolean valid = true;
		
		String clientTransport = config.getClientTransport();
		Vector transports = config.getServerTransport();
		
		Enumeration e = transports.elements();
		if(!e.hasMoreElements())
		{
			error.add(TRANSPORT_ERROR1);
			valid = false;
		}
		boolean match = false;
		while(e.hasMoreElements())
		{
		   
		    //System.out.println( (String) e.nextElement() );
		    //System.out.println( clientTransport);
		    if(((String) e.nextElement()).equals(clientTransport))
			{
				match = true;
				break;
			}
		   
		}
		if(!match)
		{
			error.add(TRANSPORT_ERROR2);
			valid = false;
		}
		return valid;
	}

	/*********************************************************
	* Checks the data types for errors. 
	*
	* @return boolean true if errors, false otherwise
	*********************************************************/
	public boolean checkExceptions()
	{
		if(config.isUserException() && !config.isSystemException())
		{
			error.add(EXCEPTION_ERROR1);
			return false;
		}
		return true;
	}

}
