/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

package mqc.protocols;

/**
 * @author Kim Christian Swenson
 */
public class CDROutputStream
{
   private boolean bigEndian;
   private byte[] buffer;
   private int bufferPtr;
   private boolean m_align;

   private static final int RESERVED_FLUFF = 1024; // extra space for security parameters
   private static final int RESERVED_GROWTH = 64; // extra growth

   public static void writeInt( byte[] buf, int startIndex, int value,  boolean bigEndian )
   {
      if(bigEndian)
      {
         buf[startIndex++] = (byte) (value >> 24);
         buf[startIndex++] = (byte) (value >> 16);
         buf[startIndex++] = (byte) (value >> 8);
         buf[startIndex++] = (byte)  value;
      }
      else
      {
         buf[startIndex++] = (byte)  value;
         buf[startIndex++] = (byte) (value >> 8);
         buf[startIndex++] = (byte) (value >> 16);
         buf[startIndex++] = (byte) (value >> 24);
      }
   }

   /**
    * Default constructor.
    */
   public CDROutputStream()
   {
      this(0, 0, false);
   }

   /**
    * Constructor. Initializes variables.
    *
    * @param n length of underlying buffer
    * @param offset  start offset in CDR stream
    * @param bigEndian  true if Big Endian encoding should be used, 
    *                   false otherwise
    */
   public CDROutputStream(int n, int offset, boolean _bigEndian)
   {
      buffer = new byte[n];
      bufferPtr = offset;
      bigEndian = _bigEndian;
      m_align = true;
   }

   /**
    * Constructor. Initializes variables.
    *
    * @param buffer  underlying buffer
    * @param offset  start offset in CDR stream
    * @param bigEndian  true if Big Endian encoding should be used, 
    *                   false otherwise
    */
   public CDROutputStream(byte[] _buffer, int offset, boolean _bigEndian)
   {
      buffer = _buffer;
      bufferPtr = offset;
      bigEndian = _bigEndian;
      m_align = true;
   }

  /**
   * @return  whether this CDR stream uses BIG ENDIAN when unmarshalling
   */
   public boolean isBigEndian()
   {
      return bigEndian;
   }

   public void setBigEndian( boolean big )
   {
      bigEndian = big;
   }

   public int bytesStreamed()
   {
      return bufferPtr;
   }

   public void setPtr( int ptr )
   {
      bufferPtr = ptr;
   }

   public boolean isAlign()
   {
      return m_align;
   }

   public void setAlign( boolean al )
   {
      m_align = al;
   }

   /**
    * Marshals a byte to the stream.
    *
    * @param octet  byte written as next octet on stream
    */
   public void writeOctet(byte octet)
   {
      bufferResizer(1);
      buffer[bufferPtr++] = octet;
   }

   /**
    * Marshals a boolean to the stream.
    *
    * @param value  boolean written as next octet on stream
    */
   public void writeBoolean(boolean value)
   {
      bufferResizer(1);
      buffer[bufferPtr++] = (byte) (value ? 1 : 0);
   }

   /**
    * Marshals a char to the stream.
    *
    * @param value  char written as next octet on stream
    */
   public void writeChar(char value)
   {
      bufferResizer(1);
      buffer[bufferPtr++] = (byte) value;
   }

   /**
    * Marshals a unsigned short to the stream.
    *
    * @param value  unsigned short written as next two octets on stream
    */
   public void writeUnsignedShort(short value)
   {
      writeShort(value);
   }

   /**
    * Marshals a short to the stream.
    *
    * @param value  signed short written as next two octets on stream
    */
   public void writeShort(short value)
   {
      //Align bufferPtr so it points to the next multiple of 2.
      bufferPtr = align(bufferPtr, 2);
      bufferResizer(2);
      if(bigEndian)
      {
         buffer[bufferPtr++] = (byte) (value >> 8);
         buffer[bufferPtr++] = (byte) value;
      }
      else
      {
         buffer[bufferPtr++] = (byte) (value);
         buffer[bufferPtr++] = (byte) (value >> 8);
      }
   }

   /**
    * Marshals a unsigned long to the stream.
    *
    * @param value  unsigned long written as next four octets on stream
    */
   public void writeUnsignedLong(int value)
   {
      writeLong(value);
   }

   /**
    * Marshals a long to the stream.
    *
    * @param value  signed long written as next four octets on stream
    */
   public void writeLong(int value)
   {
      //Align bufferPtr so it points to the next multiple of 4.
      bufferPtr = align(bufferPtr, 4);
      bufferResizer(4);
      if(bigEndian)
      {
         buffer[bufferPtr++] = (byte) (value >> 24);
         buffer[bufferPtr++] = (byte) (value >> 16);
         buffer[bufferPtr++] = (byte) (value >> 8);
         buffer[bufferPtr++] = (byte) value;
      }
      else
      {
         buffer[bufferPtr++] = (byte) (value >> 0);
         buffer[bufferPtr++] = (byte) (value >> 8);
         buffer[bufferPtr++] = (byte) (value >> 16);
         buffer[bufferPtr++] = (byte) (value >> 24);
      }
   }

   /**
    * Marshals an unsigned long long to the stream.
    *
    * @param value  unsigned long long written as next eight octets on stream
    */
   public void writeUnsignedLongLong(long value)
   {
      writeLongLong(value);
   }

   /**
    * Marshals a long long to the stream.
    *
    * @param value  signed long long written as next eight octets on stream
    */
   public void writeLongLong(long value)
   {
      //Align bufferPtr so it points to the next multiple of 8.
      bufferPtr = align(bufferPtr, 8);
      bufferResizer(8);
      if(bigEndian)
      {
         buffer[bufferPtr++] = (byte) (value >> 56);
         buffer[bufferPtr++] = (byte) (value >> 48);
         buffer[bufferPtr++] = (byte) (value >> 40);
         buffer[bufferPtr++] = (byte) (value >> 32);
         buffer[bufferPtr++] = (byte) (value >> 24);
         buffer[bufferPtr++] = (byte) (value >> 16);
         buffer[bufferPtr++] = (byte) (value >>  8);
         buffer[bufferPtr++] = (byte) value;
      }
      else
      {
         buffer[bufferPtr++] = (byte) (value >>  0);
         buffer[bufferPtr++] = (byte) (value >>  8);
         buffer[bufferPtr++] = (byte) (value >> 16);
         buffer[bufferPtr++] = (byte) (value >> 24);
         buffer[bufferPtr++] = (byte) (value >> 32);
         buffer[bufferPtr++] = (byte) (value >> 40);
         buffer[bufferPtr++] = (byte) (value >> 48);
         buffer[bufferPtr++] = (byte) (value >> 56);
      }
   }

   /**
    * Marshals a float to the stream.
    *
    * @param value  float written as next four octets on stream
    */
   //public void writeFloat(float value)
   //{
      //writeLong(java.lang.Float.floatToIntBits(value));
   //}

   /**
    * Marshals a double to the stream.
    *
    * @param value  double written as next eight octets on stream
    */
   //public void write_double(double value)
   //{
      //writeLongLong(java.lang.Double.doubleToLongBits(value));
   //}

   /**
    * Marshals a string to the stream.
    *
    * @param value  string written next on stream
    */
   public void writeString(String value)
   {
      int len = value.length() + 1;
      writeLong(len);
      bufferResizer(len);
      System.arraycopy(value.getBytes(), 0, buffer, bufferPtr, len-1);
      bufferPtr += len-1;
      writeOctet((byte)0);
   }

   public void writeOctetSequence( byte[] seq )
   {
      bufferResizer( 4 + seq.length );
      writeLong( seq.length );
      System.arraycopy(seq, 0, buffer, bufferPtr, seq.length);
      bufferPtr += seq.length;
   }

   /**
    * Resizes the buffer if it's too small.
    *
    * @param size  the minimum length in bytes that must be available
    */
   private void bufferResizer(int size)
   {
      if(mqc.Config.intPacketSize == 0 && buffer.length < (bufferPtr + size + RESERVED_FLUFF ) )
      {
         // add 32 extra bytes in case of alignment needs.
         byte[] tmp = new byte[buffer.length + size + RESERVED_GROWTH + RESERVED_FLUFF];
         System.arraycopy(buffer, 0, tmp, 0, bufferPtr);
         buffer = tmp;
      }
   }

   /**
    * Sets the buffer. This method automatically resets the CDR stream.
    *
    * @param buffer  underlying buffer used to hold the CDR stream
    *
    */
   public void setBuffer(byte[] buffer)
   {
      this.buffer = buffer;
   }

   /**
    * Gets the buffer.
    *
    * @return  underlying buffer used to hold the CDR stream
    *
    */
   public byte[] getBuffer()
   {
      return buffer;
   }

   /**
    * Resets this CDR stream causing it to start writing at the 
    * beginning of the stream.
    */
   public void reset()
   {
      bufferPtr = 0;
   }

   private int align( int ptr, int alignment )
   {
      if ( m_align )
      {
         return (ptr + alignment - 1) & ~(alignment-1);
      }
      else
      {
         return ptr;
      }
   }
}

