/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

include(`Config_macros.m4')

package mqc.transports;

import mqc.*;
import mqc.ior.*;
import mqc.holders.*;
import mqc.udp.*;
import java.io.*;
import java.util.*;
import java.net.*;
__VALUERED__(
import mqc.util.CRC32;
)

/**
* UDP transport class
*
* @author Olav Haugan
* @author Tarana Damania
*/
public class Stop_N_Wait_UDPTransport extends Transport
{
   private String host;
   private int port;
   private int localPort;
   private UDPClient client;
   private UDPSocket udpSocket;
   private int protocol;
   __VALUERED__(
   private CRC32 chkSum = null;
   )
   //????
   private static final int MAX_DATAGRAM_SIZE = 1024;

   /**
    * Constructor. Initializes variables.
    */
   public Stop_N_Wait_UDPTransport()
   {
      super();
      host = null;
      port = 0;
      localPort = 0;
      client = null;
      __VALUERED__(
      chkSum = new CRC32();
      )
   }

   /**
   * Constructor. Initializes variables.
   *
   * @param mqc.Delegate The delegate object
   */
   public Stop_N_Wait_UDPTransport(mqc.Delegate delegate)
   {
      super();
      Profile profile = delegate.ior.getProfile();
      protocol = delegate.protocol;
      TaggedProfile tagProfile = profile.getTaggedProfile(TaggedProfile.TAG_INTERNET_UDP);
      host = tagProfile.host;
      port = tagProfile.port;
      __VALUERED__(
      chkSum = new CRC32();
      )
      init();
   }

   /**
   * Initializes a connection.
   * We'll connect to the server here.
   *
   * @return boolean true if socket created successfully, false otherwise
   */
   public boolean init()
   {
      try
      {
         DatagramSocket clientSocket;

         //Create the UDP Client
         //---------------------------
         client =  new UDPClient();

         //Get the local Port
         //-------------------------
         localPort = client.localPort;

         //Connect to the server
         //---------------------------
         clientSocket = client.connect( host, port);

         if ( clientSocket != null )
         {
            //Create a Go Back N Socket
            //-----------------------------
            udpSocket = new UDPStopAndWaitSocket( clientSocket, host, client.server_com_port );
         }
      }
      catch(Exception e)
      {
         System.out.println("Error creating socket");
         return false;
      }
      return true;
   }

   /**
   * Sends a message
   *
   * @param byte[] The buffer to send
   * @param int The number of bytes to transmit
   * @return boolean true if data is sent successfully, false otherwise
   */
   public boolean send(byte[] buffer, int length)
   {
      try
      {
         __VALUERED__(
         // calculate a CRC32 checksum for the byte array
         byte[] buffActLen = new byte[Config.intPacketSize - ORB.CKSUM_LEN];
         System.arraycopy(buffer,0,buffActLen,0,length);
         int intCheckSum = chkSum.crc32(buffActLen);
         __DEBUG__( System.out.println("Checksum is: " + intCheckSum);)

         // write the integer checksum value into a byte array (big endian fmt)
         byte[] bCheckSum = new byte[ORB.CKSUM_LEN];
         bCheckSum[0] = (byte) (intCheckSum >> 24);
         bCheckSum[1] = (byte) (intCheckSum >> 16);
         bCheckSum[2] = (byte) (intCheckSum >> 8);
         bCheckSum[3] = (byte)  intCheckSum;

         // append the checksum byte array to the front of the message
         byte[] buffTemp = new byte[bCheckSum.length + buffer.length];
         System.arraycopy(bCheckSum,0,buffTemp,0,bCheckSum.length);
         System.arraycopy(buffer,0,buffTemp,bCheckSum.length,buffer.length);
         )

         __TEMPORAL__(
         // here we can use the temporal variable in the Config file to determine
         //  what the level of temporal redundancy is
         for(int i=0;i<Config.temporal;i++)
         {
            __VALUERED__(
            // here we need to load the message into a fixed length packet
            byte[] bMessagePacket = new byte[Config.intPacketSize];
            // copy the message into the packet array
            System.arraycopy(buffTemp,0,bMessagePacket,0,length + bCheckSum.length);
            // now send the fixed length packet
            udpSocket.send(bMessagePacket,bMessagePacket.length);
            //udpSocket.send(buffTemp,length + bCheckSum.length);
            )
            __NOVALUERED__(
            // here we need to load the message into a fixed length packet
            byte[] bMessagePacket = new byte[Config.intPacketSize];
            // copy the message into the packet array
            System.arraycopy(buffer,0,bMessagePacket,0,length);
            // now send the fixed length packet
            udpSocket.send(bMessagePacket,bMessagePacket.length);
            //udpSocket.send(buffer, length);
            )
         }
         )
         __NOTEMPORAL__(
         __VALUERED__(
         // here we need to load the message into a fixed length packet
         byte[] bMessagePacket = new byte[Config.intPacketSize];
         // copy the message into the packet array
         System.arraycopy(buffTemp,0,bMessagePacket,0,length + bCheckSum.length);
         // now send the fixed length packet
         udpSocket.send(bMessagePacket,bMessagePacket.length);
         //udpSocket.send(buffTemp,length + bCheckSum.length);
         )
         __NOVALUERED__(
         // here we need to load the message into a fixed length packet
         byte[] bMessagePacket = new byte[Config.intPacketSize];
         // copy the message into the packet array
         System.arraycopy(buffer,0,bMessagePacket,0,length);
         // now send the fixed length packet
         udpSocket.send(bMessagePacket,bMessagePacket.length);
         //udpSocket.send(buffer, length);
         )
         )
      }
      catch(Exception e)
      {
         System.out.println("Error sending data to socket");
         return false;
      }
      return true;
   }

   /**
   * Waits for a message and receive it.
   *
   * @param Message The message sent
   * @param BooleanHolder true if big endian, false otherwise
   * @return byte[] The buffer received
   */
   public byte[] waitForReply(Message msg, BooleanHolder bigEndian)
   {
      byte[] data = new byte[Config.intPacketSize];

      int intPacketSize = udpSocket.receive(data);

      DataInputStream in = new DataInputStream(new ByteArrayInputStream(data, 0, intPacketSize));
      byte[] buffer = msg.deMarshal(in, bigEndian);
      return buffer;
   }
}
